#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# License: MIT


import glob
import os

from . import ADJOINT, dolfin
from .complex import *
from .formulation import Maxwell2D
from .geometry import *
from .materials import *
from .simulation import Simulation
from .source import *


class BoxPML2D(Geometry):
    def __init__(
        self,
        box_size=(1, 1),
        box_center=(0, 0),
        pml_width=(0.2, 0.2),
        model_name="2D box with PMLs",
        mesh_name="mesh.msh",
        data_dir=None,
        **kwargs,
    ):
        super().__init__(
            model_name=model_name,
            mesh_name=mesh_name,
            data_dir=data_dir,
            dim=2,
            **kwargs,
        )
        self.box_size = box_size
        self.box_center = box_center
        self.pml_width = pml_width

        def _addrect_center(rect_size):
            corner = -np.array(rect_size) / 2
            corner = tuple(corner) + (0,)
            return self.add_rectangle(*corner, *rect_size)

        def _translate(tag, t):
            translation = tuple(t) + (0,)
            self.translate(self.dimtag(tag), *translation)

        def _add_pml(s, t):
            pml = _addrect_center(s)
            _translate(pml, t)
            return pml

        box = _addrect_center(self.box_size)
        s = (self.pml_width[0], self.box_size[1])
        t = np.array([self.pml_width[0] / 2 + self.box_size[0] / 2, 0])
        pmlxp = _add_pml(s, t)
        pmlxm = _add_pml(s, -t)
        s = (self.box_size[0], self.pml_width[1])
        t = np.array([0, self.pml_width[1] / 2 + self.box_size[1] / 2])
        pmlyp = _add_pml(s, t)
        pmlym = _add_pml(s, -t)

        s = (self.pml_width[0], self.pml_width[1])
        t = np.array(
            [
                self.pml_width[0] / 2 + self.box_size[0] / 2,
                self.pml_width[1] / 2 + self.box_size[1] / 2,
            ]
        )
        pmlxypp = _add_pml(s, t)
        pmlxymm = _add_pml(s, -t)
        pmlxypm = _add_pml(s, (-t[0], t[1]))
        pmlxymp = _add_pml(s, (t[0], -t[1]))

        all_dom = [
            box,
            pmlxp,
            pmlxm,
            pmlyp,
            pmlym,
            pmlxypp,
            pmlxypm,
            pmlxymm,
            pmlxymp,
        ]
        _translate(all_dom, self.box_center)

        self.box = box
        self.pmls = all_dom[1:]

        self.fragment(self.box, self.pmls)
        self.add_physical(box, "box")
        self.add_physical([pmlxp, pmlxm], "pmlx")
        self.add_physical([pmlyp, pmlym], "pmly")
        self.add_physical([pmlxypp, pmlxypm, pmlxymm, pmlxymp], "pmlxy")


class Scatt2D(Simulation):
    def __init__(
        self,
        geometry,
        epsilon,
        mu,
        source,
        boundary_conditions={},
        polarization="TE",
        degree=1,
        mat_degree=1,
        pml_stretch=1 - 1j,
    ):
        assert isinstance(geometry, BoxPML2D)
        assert source.dim == 2
        function_space = ComplexFunctionSpace(geometry.mesh, "CG", degree)
        pmlx = PML(
            "x", stretch=pml_stretch, matched_domain="box", applied_domain="pmlx"
        )
        pmly = PML(
            "y", stretch=pml_stretch, matched_domain="box", applied_domain="pmly"
        )
        pmlxy = PML(
            "xy", stretch=pml_stretch, matched_domain="box", applied_domain="pmlxy"
        )

        epsilon_coeff = Coefficient(
            epsilon, geometry, pmls=[pmlx, pmly, pmlxy], degree=mat_degree
        )
        mu_coeff = Coefficient(
            mu, geometry, pmls=[pmlx, pmly, pmlxy], degree=mat_degree
        )

        coefficients = epsilon_coeff, mu_coeff
        no_source_domains = ["box", "pmlx", "pmly", "pmlxy"]
        source_domains = [
            dom for dom in geometry.domains if dom not in no_source_domains
        ]
        formulation = Maxwell2D(
            geometry,
            coefficients,
            function_space,
            source=source,
            source_domains=source_domains,
            reference="box",
            polarization=polarization,
            boundary_conditions=boundary_conditions,
        )

        super().__init__(geometry, formulation)

        self.degree = degree

    def solve_system(self, again=False):
        u = super().solve_system(again=again, vector_function=False)
        self.solution = {}
        self.solution["diffracted"] = u
        self.solution["total"] = u + self.source.expression
        return u

    # def local_density_of_states(self, x, y):
    #     ldos = np.zeros((len(x), len(y)))
    #     for ix, x_ in enumerate(x):
    #         for iy, y_ in enumerate(y):
    #             print(x_, y_)
    #             self.source.position = x_, y_
    #             self.assemble_rhs()
    #             u = self.solve_system(again=True)
    #             ldos[ix, iy] = u(self.source.position).imag
    #     return ldos

    def plot_field(
        self,
        ax=None,
        mincmap=None,
        maxcmap=None,
        fig=None,
        phase=0,
        callback=None,
        **kwargs,
    ):

        import matplotlib as mpl

        from .plot import plt

        u = self.solution["total"]
        if ax == None:
            ax = plt.gca()
        if "cmap" not in kwargs:
            kwargs["cmap"] = "RdBu_r"

        f = u * phase_shift(phase, degree=self.degree)
        fplot = f.real
        if ADJOINT:
            fplot = project(fplot, self.formulation.real_function_space)
        pp = dolfin.plot(fplot, **kwargs)

        ppmax = pp.cvalues[-1]
        ppmin = pp.cvalues[0]
        ax.set_aspect(1)
        mincmap = mincmap or ppmin
        maxcmap = maxcmap or ppmax
        pp.set_clim(mincmap, maxcmap)

        cm = plt.cm.ScalarMappable(cmap=kwargs["cmap"])
        cm.set_clim(mincmap, maxcmap)

        fig = plt.gcf() if fig is None else fig
        cb = fig.colorbar(cm, ax=ax)

        if callback is not None:
            callback(**kwargs)

        return pp, cb

    def animate_field(self, n=11, filename="animation.gif", **kwargs):
        import tempfile

        from PIL import Image

        from .plot import plt

        anim = []
        tmpdir = tempfile.mkdtemp()
        fp_in = f"{tmpdir}/animation_tmp_*.png"
        phase = np.linspace(0, 2 * np.pi, n + 1)[:n]
        for iplot in range(n):
            number_str = str(iplot).zfill(4)
            pngname = f"{tmpdir}/animation_tmp_{number_str}.png"
            p = self.plot_field(anim=phase[iplot], **kwargs)
            fig = plt.gcf()
            fig.savefig(pngname)
            fig.clear()
            anim.append(p)

        plt.close(fig)

        img, *imgs = [Image.open(f) for f in sorted(glob.glob(fp_in))]
        img.save(
            fp=filename,
            format="GIF",
            append_images=imgs,
            save_all=True,
            duration=200,
            loop=0,
        )
        os.system(f"rm -f {tmpdir}/animation_tmp_*.png")
        return anim



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# License: MIT

import tempfile

import meshio
import numpy as np

from . import dolfin


def read_mesh(mesh_file, data_dir=None, dim=3):
    meshio_mesh = meshio.read(mesh_file)

    points = meshio_mesh.points[:, :2] if dim == 2 else meshio_mesh.points
    physicals = meshio_mesh.cell_data_dict["gmsh:physical"]

    cell_types, data_gmsh = zip(*physicals.items())
    cells = {ct: [] for ct in cell_types}

    for cell_type in cell_types:
        for cell in meshio_mesh.cells:
            if cell.type == cell_type:
                cells[cell_type].append(cell.data)
        cells[cell_type] = np.vstack(cells[cell_type])

    mesh_data = {}
    for cell_type, data in zip(cell_types, data_gmsh):
        meshio_data = meshio.Mesh(
            points=points,
            cells={cell_type: cells[cell_type]},
            cell_data={cell_type: [data]},
        )
        meshio.xdmf.write(f"{data_dir}/{cell_type}.xdmf", meshio_data)
        mesh_data[cell_type] = meshio_data

    dolfin_mesh = dolfin.Mesh()

    cell_type = "tetra" if dim == 3 else "triangle"
    with dolfin.XDMFFile(f"{data_dir}/{cell_type}.xdmf") as infile:
        infile.read(dolfin_mesh)

    markers = {}

    dim_map = dict(line=1, triangle=2, tetra=3)
    for cell_type in cell_types:
        mvc = dolfin.MeshValueCollection("size_t", dolfin_mesh, dim_map[cell_type])
        with dolfin.XDMFFile(f"{data_dir}/{cell_type}.xdmf") as infile:
            infile.read(mvc, cell_type)
        markers[cell_type] = dolfin.cpp.mesh.MeshFunctionSizet(dolfin_mesh, mvc)

    return dict(mesh=dolfin_mesh, markers=markers)


class MarkedMesh(object):
    def __init__(self, filename, geometric_dimension=3, data_dir=None):
        self.data_dir = data_dir
        self.geometric_dimension = geometric_dimension
        self.filename = filename

        data_dir = data_dir or tempfile.mkdtemp()
        dic = read_mesh(filename, dim=geometric_dimension, data_dir=data_dir)
        self.mesh = dic["mesh"]
        self.markers = dic["markers"]
        self.dimension = self.mesh.geometric_dimension()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# License: MIT

import copy

import matplotlib.pyplot as plt
import numpy as np
from dolfin.common.plotting import mesh2triang
from matplotlib.tri import Triangulation

from gyptis.complex import *

from . import dolfin


def get_boundaries(markers, domain=None, shift=(0, 0)):

    data = markers.array()
    triang = mesh2triang(markers.mesh())
    if domain == None:
        ids = np.unique(data)
    else:
        ids = [domain]

    triangulations = []
    for id in ids:
        triang_ = copy.deepcopy(triang)
        triang_.set_mask(data != id)
        triangulations.append(triang_)

    sub_bnds = []
    for triangtest in triangulations:
        maskedTris = triangtest.get_masked_triangles()
        verts = np.stack((triangtest.x[maskedTris], triangtest.y[maskedTris]), axis=-1)
        all_vert = np.vstack(verts).T
        sub_triang = Triangulation(*all_vert)

        boundaries = []
        for i in range(len(sub_triang.triangles)):
            for j in range(3):
                if sub_triang.neighbors[i, j] < 0:
                    boundaries.append(
                        (
                            sub_triang.triangles[i, j],
                            sub_triang.triangles[i, (j + 1) % 3],
                        )
                    )
        boundaries = np.asarray(boundaries)

        bndpnts = (
            shift[0] + sub_triang.x[boundaries].T,
            shift[1] + sub_triang.y[boundaries].T,
        )

        sub_bnds.append(bndpnts)

    return sub_bnds


def plot_boundaries(markers, domain=None, shift=(0, 0), ax=None, **kwargs):
    sub_bnds = get_boundaries(markers, domain=domain, shift=shift)
    if "c" not in kwargs and "color" not in kwargs:
        kwargs["color"] = "k"
    if ax is None:
        ax = plt.gca()
    l = []
    for bndpnts in sub_bnds:
        l_ = ax.plot(*bndpnts, **kwargs)
        l.append(l_)
    return l


# def plot_subdomains(markers, alpha=0.3):
#     a = dolfin.plot(markers, cmap="binary", alpha=alpha, lw=0.00, edgecolor="face")
#     return a
#     # a.set_edgecolors((0.1, 0.2, 0.5, 0.))


def plot_subdomains(markers, **kwargs):
    return plot_boundaries(markers, **kwargs)


def plotcplx(test, ax=None, markers=None, proj_space=None, ref_cbar=False, **kwargs):
    if ax is None:
        fig, ax = plt.subplots(1, 2)

    if "cmap" not in kwargs:
        kwargs["cmap"] = "RdBu_r"
    proj = proj_space is not None

    if proj:
        test = project(test, proj_space)
    P, C = [], []
    for a, t in zip(ax, [test.real, test.imag]):
        plt.sca(a)
        p = dolfin.plot(t, **kwargs)
        cbar = plt.colorbar(p)
        if markers:
            plot_subdomains(markers, **kwargs)
        if ref_cbar:
            v = test.real.vector().get_local()
            mn, mx = min(v), max(v)
            md = 0.5 * (mx + mn)
            cbar.set_ticks([mn, md, mx])
            lab = [f"{m:.2e}" for m in [mn, md, mx]]
            cbar.set_ticklabels(lab)
        P.append(p)
        C.append(cbar)
    return P, C


def plot(test, ax=None, markers=None, proj_space=None, **kwargs):
    proj = proj_space is not None
    if "cmap" not in kwargs:
        kwargs["cmap"] = "inferno"
    if ax is None:
        fig, ax = plt.subplots(1, 1)
    if proj:
        test = project(test, proj_space)
    plt.sca(ax)
    p = dolfin.plot(test, **kwargs)
    cbar = plt.colorbar(p)
    if markers:
        plot_subdomains(markers)

    return p, cbar



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# License: MIT

from . import dolfin
from .complex import Complex, assemble


class Simulation:
    def __init__(self, geometry, formulation=None):
        self.geometry = geometry
        self.formulation = formulation
        self.coefficients = formulation.coefficients
        self.function_space = formulation.function_space
        self._source = formulation.source
        self.boundary_conditions = formulation.boundary_conditions
        self.mesh = self.geometry.mesh
        self._boundary_conditions = []
        self.dx = formulation.dx
        self.ds = formulation.ds
        self.dS = formulation.dS

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, value):
        self._source = value

    def assemble_lhs(self, **kwargs):
        self.formulation.build_lhs(**kwargs)
        self.matrix = assemble(self.formulation.lhs)
        return self.matrix

    def assemble_rhs(self, **kwargs):
        self.formulation.build_rhs(**kwargs)
        self.vector = assemble(self.formulation.rhs)
        return self.vector

    def assemble(self, **kwargs):
        self.matrix = self.assemble_lhs(**kwargs)
        self.vector = self.assemble_rhs(**kwargs)
        return self.matrix, self.vector

    def apply_boundary_conditions(self):
        bcs = self.formulation.build_boundary_conditions()
        for bc in bcs:
            bc.apply(self.matrix, self.vector)

    def solve_system(self, again=False, vector_function=True):
        if vector_function:
            element = self.function_space.split()[0].ufl_element()
            V_vect = dolfin.VectorFunctionSpace(
                self.mesh, element.family(), element.degree()
            )
            u = dolfin.Function(V_vect)
        else:
            u = dolfin.Function(self.function_space)

        if not again:
            self.solver = dolfin.LUSolver(self.matrix, "mumps")

        self.solver.solve(u.vector(), self.vector)
        solution = Complex(*u.split())
        return solution

    def solve(self, again=False, **kwargs):
        self.assemble(**kwargs)
        self.apply_boundary_conditions()
        return self.solve_system(again=again, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# License: MIT


from . import dolfin


class _DirichletBC(dolfin.DirichletBC):
    def __init__(self, *args, **kwargs):
        self.subdomain_dict = args[-1]
        if not callable(args[2]):
            args = list(args)
            args[-2] = self.subdomain_dict[args[-2]]
            args = tuple(args[:-1])
        super().__init__(*args)


class DirichletBC:
    def __new__(self, *args, **kwargs):
        W = args[0]
        value = args[1]
        Wre, Wim = W.split()
        bcre = _DirichletBC(Wre, value.real, *args[2:], **kwargs)
        bcim = _DirichletBC(Wim, value.imag, *args[2:], **kwargs)
        return bcre, bcim


class PeriodicBoundary2DX(dolfin.SubDomain):
    def __init__(self, period, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.period = period

    def inside(self, x, on_boundary):
        return bool(dolfin.near(x[0], -self.period / 2) and on_boundary)

    # # Left boundary is "target domain" G
    # def inside(self, x, on_boundary):
    #     return bool(
    #         x[0] - self.period / 2 < dolfin.DOLFIN_EPS
    #         and x[0] - self.period / 2 > -dolfin.DOLFIN_EPS
    #         and on_boundary
    #     )

    def map(self, x, y):
        y[0] = x[0] - self.period
        y[1] = x[1]

    #
    # def map(self, x, y):
    #     if dolfin.near(x[0], self.period / 2):
    #         y[0] = x[0] - self.period
    #         y[1] = x[1]
    #     else:
    #         y[0] = -1000
    #         y[1] = -1000


class BiPeriodicBoundary3D(dolfin.SubDomain):
    def __init__(self, period, **kwargs):
        self.period = period
        super().__init__(**kwargs)

    def inside(self, x, on_boundary):
        return bool(
            (
                dolfin.near(x[0], -self.period[0] / 2)
                or dolfin.near(x[1], -self.period[1] / 2)
            )
            and (
                not (
                    (
                        dolfin.near(x[0], -self.period[0] / 2)
                        and dolfin.near(x[1], self.period[1] / 2)
                    )
                    or (
                        dolfin.near(x[0], self.period[0] / 2)
                        and dolfin.near(x[1], -self.period[1] / 2)
                    )
                )
            )
            and on_boundary
        )

    def map(self, x, y):
        if dolfin.near(x[0], self.period[0] / 2) and dolfin.near(
            x[1], self.period[1] / 2
        ):
            y[0] = x[0] - self.period[0]
            y[1] = x[1] - self.period[1]
            y[2] = x[2]
        elif dolfin.near(x[0], self.period[0] / 2):
            y[0] = x[0] - self.period[0]
            y[1] = x[1]
            y[2] = x[2]
        elif dolfin.near(x[1], self.period[1] / 2):
            y[0] = x[0]
            y[1] = x[1] - self.period[1]
            y[2] = x[2]
        else:
            y[0] = -1000  # -self.period[0]*2.
            y[1] = -1000  # -self.period[1]*2.
            y[2] = -1000  # 0.


def prepare_boundary_conditions(bc_dict):
    valid_bcs = ["PEC"]
    boundary_conditions = []
    pec_bnds = []
    for bnd, cond in bc_dict.items():
        if cond not in valid_bcs:
            raise ValueError(f"Unknown boundary condition {cond}")
        else:
            pec_bnds.append(bnd)
    return pec_bnds


def build_pec_boundary_conditions(pec_bnds, geometry, function_space, applied_function):
    boundary_conditions = []
    for bnd in pec_bnds:
        bc = DirichletBC(
            function_space,
            applied_function,
            geometry.boundary_markers,
            bnd,
            geometry.boundaries,
        )
        [boundary_conditions.append(b) for b in bc]
    return boundary_conditions



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# License: MIT


__all__ = [
    "Complex",
    "Layered",
    "Scattering",
    "Grating",
    "PlaneWave",
    "LineSource",
    "BoxPML",
    "Layered",
]

#
# from . import complex
# from . import complex
from .complex import Complex
from .grating2d import Grating2D, Layered2D
from .grating3d import Grating3D, Layered3D
from .scattering2d import BoxPML2D, Scatt2D
from .scattering3d import BoxPML3D, Scatt3D
from .source import LineSource, PlaneWave


class BoxPML:
    """BoxPML(dim, x,y,z)
    Creates a computational domain with Perfectly Matched Layers (PMLs).

    Parameters
    ----------
    dim : int
        Geometric dimension (either 2 or 3, the default is 3).
    *args : type
        Description of parameter `*args`.
    **kwargs : type
        Description of parameter `**kwargs`.

    Returns
    -------
    type
        Description of returned object.
    """

    def __new__(self, dim=3, *args, **kwargs):

        if dim not in [2, 3]:
            raise ValueError("dimension must be 2 or 3")
        if dim == 3:
            return BoxPML3D(*args, **kwargs)
        else:
            return BoxPML2D(*args, **kwargs)


class Layered:
    def __new__(self, dim=3, *args, **kwargs):
        if dim not in [2, 3]:
            raise ValueError("dimension must be 2 or 3")
        if dim == 3:
            return Layered3D(*args, **kwargs)
        else:
            return Layered2D(*args, **kwargs)


class Scattering:
    def __new__(self, *args, **kwargs):
        geom = kwargs.get("geometry") or args[0]
        if geom.dim not in [2, 3]:
            raise ValueError("dimension must be 2 or 3")
        if geom.dim == 3:
            return Scatt3D(*args, **kwargs)
        else:
            return Scatt2D(*args, **kwargs)


class Grating:
    def __new__(self, *args, **kwargs):
        geom = kwargs.get("geometry") or args[0]
        if geom.dim not in [2, 3]:
            raise ValueError("dimension must be 2 or 3")
        if geom.dim == 3:
            return Grating3D(*args, **kwargs)
        else:
            return Grating2D(*args, **kwargs)
